import java.util.ArrayList;
import java.util.Iterator;

/**
 * abstract class for  display *
 * @param <A> A subclass
 */
public abstract class GraphNode<A extends GraphNode> {

	static BreadthFirstSearch bfs = new BreadthFirstSearch();
	public static int circleDepths[];

	private double calculatedSize = -1;

	protected ArrayList<A> contacts;

	private double currentDepth = 0;

	private double currentAngle = 0;

	private double goalDepth = 0;

	private double goalAngle = 0;

	private double oldAngle = 0;

	private double oldDepth = 0;

	private double dt = 0;

	private double t = 0;

	private int animationSteps = 0;

	private boolean positionCalculated = false;

	private boolean highlighted = false;

	private static final double ANGLESTEP = 5.0;

	private static final double DEPTHSTEP = 5.0;
	
	private static final int MIN_ANIMATIONSTEPS = 10;

	enum Direction {
		DIRECT, PLUS360, MINUS360
	};

	private Direction animationDirection;

	public abstract int getUserSize();

	/**
	 * has to be called before anything is calculated
	 *
	 */
	public void resetCalculated() {
		positionCalculated = false;
		for (A contact : contacts) {

			if (contact.getGraphDepth() <= getGraphDepth())
				continue;
			contact.resetCalculated();
		}
	}

	public boolean isHighlighted() {
		return highlighted;
	}

	public void setHighlighted(boolean value) {
		highlighted = value;
	}

	public double getCalculatedSize() {
		if (calculatedSize < 0)
			getAngularWidth(null);
		return calculatedSize;
	}

	public Iterator<A> getNeighbours() {
		return contacts.iterator();
	}

	public void startBFS() {
		// creates new object and starts search
		bfs = new BreadthFirstSearch(this);
	}

	public int getGraphDepth() {
		return bfs.getSearchResult().get(this);
	}

	/**
	 * stops all animation for this node.
	 *
	 */
	public void finishAnimation() {
		oldAngle = currentAngle;
		currentAngle = goalAngle;
		currentDepth = goalDepth;
		currentAngle = (currentAngle + 360) % 360;
	}
	/**
	 * starts an animation of this object
	 * @param goalDepth
	 * @param goalAngle
	 */
	public void animateTo(double goalDepth, double goalAngle) {
		oldAngle = currentAngle;
		oldDepth = currentDepth;
		this.goalDepth = goalDepth;
		this.goalAngle = goalAngle;

		int nrAngleSteps;
		int nrDepthSteps = (int) (Math.abs(goalDepth - currentDepth) / DEPTHSTEP);

		double diffAngle = goalAngle - currentAngle;

		// There are four different cases for goalAngle and currentAngle
		// 1) The direct way is shortest, goalAngle > currentAngle
		// 2) The direct way is shortest, currentAngle > goalAngle
		// 3) The way passing 360� is shorter, goalAngle > currentAngle
		// 4) The way passing 360� is shorter, currentAngle > goalAngle

		// The direct way is shorter, this is easy!
		if (Math.abs(diffAngle) < 360 - Math.abs(diffAngle)) {
			nrAngleSteps = (int) (Math.abs(diffAngle) / ANGLESTEP);		
			animationSteps = Math.max(nrAngleSteps, nrDepthSteps);

			animationDirection = Direction.DIRECT;
		}

		// The way passing 360� is shorter, we need to look at the two different
		// cases
		else {
			nrAngleSteps = (int) ((360 - Math.abs(diffAngle)) / ANGLESTEP);			
			animationSteps = Math.max(nrAngleSteps, nrDepthSteps);

			if (goalAngle >= currentAngle) {
				animationDirection = Direction.PLUS360;
			} else {
				animationDirection = Direction.MINUS360;
			}
		}
		
		animationSteps = Math.max(animationSteps, MIN_ANIMATIONSTEPS);

		t = 0.0f;
		dt = 1.0f / animationSteps;
	}

	public double getAngle() {
		return currentAngle;
	}

	public double getDepth() {
		return currentDepth;
	}

	public double getGoalAngle() {
		return goalAngle;
	}
	public double getGoalDepth() {
		return goalDepth;
	}

	/**
	 * one animation step for this node
	 * @return true if another step is needed, false if animation finished
	 */
	public boolean animationStep() {
		if (t >= 1.0f) {
			return false;
		} else if ((t + dt) >= 1.0f) {
			t = 1.0f;
			finishAnimation();
			return true;
		} else {
			if (animationDirection == Direction.DIRECT)
				currentAngle = interpolate(oldAngle, goalAngle, t);
			else if (animationDirection == Direction.PLUS360)
				currentAngle = interpolate(360 + oldAngle, goalAngle, t);
			else if (animationDirection == Direction.MINUS360)
				currentAngle = interpolate(oldAngle, 360 + goalAngle, t);

			currentDepth = interpolate(oldDepth, goalDepth, t);
			t = t + dt;

			return true;
		}
	} // interpolate between a and b, t = 0..1

	/**
	 * interpolation functions
	 * @param a lower value
	 * @param b upper value
	 * @param t time
	 * @return interpolated value
	 */
	private double interpolate(double a, double b, double t) {
		// atan is always between -pi/2 and pi/2, we want results 0..1
		// atan(-5) is approx. -pi/2 (not really, that's why we have to scale
		// those results back between 0 and 1 again) atan(5) is approx. pi/2,
		// therefore we scale t from 0..1 to -5..5

		double interval = 5;

		// scale t between -interval and +interval
		double tScaled = -interval + t * interval * 2;

		double atanAtBorder = (Math.PI / 2 + Math.atan(-interval)) / Math.PI;
		double atanScaled = (Math.PI / 2 + Math.atan(tScaled)) / Math.PI;

		atanScaled = (-atanAtBorder + atanScaled)
				* (1 / (1 - 2 * atanAtBorder));

		return a + atanScaled * (b - a);
	}

	// classical dfs search
	public double getAngularWidth(A parent) {

		double childSize = 0;
		for (A contact : contacts) {
			if (contact == parent || contact.getGraphDepth() <= getGraphDepth())
				continue;

			childSize += contact.getAngularWidth(this);
		}
		if (getGraphDepth() == 0) {
			calculatedSize = childSize;
			return calculatedSize;
		}

		calculatedSize = Math.max(getUserSize() / getGraphDepth(), childSize);
		return calculatedSize;
	}

	// distance between alpha and beta is used for visualisation
	// 0<= alpha <= beta<=360
	// calulates the position recursiv for all contacts that have a greater
	// depth
	public void calculatePosition(double alpha, double beta) {
		if (positionCalculated)
			return;
		double totalSize = getCalculatedSize();

		// Our node is in the middle of the lower angle alpha, and the upper
		// limit beta
		// little trick for more exact math: alpha + (beta-alpha) / 2
		double tmp = getGraphDepth();
		if(circleDepths != null && tmp<circleDepths.length)
			tmp = circleDepths[getGraphDepth()];
		animateTo(tmp, (alpha + beta) / 2);
		positionCalculated = true;
		double lastAngle = alpha;
		for (A contact : contacts) {

			// recursion termination
			if (contact.getGraphDepth() <= getGraphDepth())
				continue;

			// calculating the maximum angle for our node, to pass it to the
			// recursion
			double maxAngle = lastAngle
					+ (contact.getCalculatedSize() / totalSize)
					* (beta - alpha);
			contact.calculatePosition(lastAngle, maxAngle);
			lastAngle = maxAngle;
		}
	}
}
